import unittest
from pathlib import Path

import pytest

from pyjelly.integrations.generic.generic_sink import (
    IRI,
    BlankNode,
    GenericStatementSink,
    Literal,
    Triple,
)


class TestGenericStatementSink(unittest.TestCase):
    def setUp(self) -> None:
        self.test_obj = GenericStatementSink("graph_id")
        self.store_content = (
            IRI("http://example.com/s"),
            IRI("http://example.com/p"),
            IRI("http://example.com/o"),
        )
        self.test_obj.add(Triple(*self.store_content))
        self.test_obj.bind("ex", IRI("http://example.com/"))

    def test_namespaces_property(self) -> None:
        namespaces = list(self.test_obj.namespaces)
        assert namespaces[0][0] == "ex"
        assert repr(namespaces[0][1]) == repr(IRI("http://example.com/"))

    def test_identifier_property(self) -> None:
        assert self.test_obj.identifier == "graph_id"

    def test_store_property(self) -> None:
        for statement in self.test_obj.store:
            assert statement == Triple(*self.store_content)

    def test_iterator(self) -> None:
        for statement in self.test_obj:
            assert statement == Triple(*self.store_content)

    def test_len(self) -> None:
        assert len(self.test_obj) == 1


@pytest.mark.parametrize(
    (
        "input_id",
        "expected_repr",
    ),
    [
        ("bn1", "BlankNode(identifier=bn1)"),
        ("929292", "BlankNode(identifier=929292)"),
        ("utf-8~", "BlankNode(identifier=utf-8~)"),
    ],
)
def test_bn_repr(input_id: str, expected_repr: str) -> None:
    bn = BlankNode(input_id)
    assert repr(bn) == expected_repr


@pytest.mark.parametrize(
    (
        "input_id",
        "expected_repr",
    ),
    [
        ("bn1", "_:bn1"),
        ("929292", "_:929292"),
        ("utf-8~", "_:utf-8~"),
    ],
)
def test_bn_str(input_id: str, expected_repr: str) -> None:
    bn = BlankNode(input_id)
    assert str(bn) == expected_repr


@pytest.mark.parametrize(
    (
        "lex",
        "langtag",
        "datatype",
        "expected_str",
    ),
    [
        ("hello", None, None, '"hello"'),
        ("", None, None, '""'),
        ("123", None, None, '"123"'),
        ("true", None, None, '"true"'),
        ("hello", "en", None, '"hello"@en'),
        ("bonjour", "fr", None, '"bonjour"@fr'),
        ("hallo", "de", None, '"hallo"@de'),
        ("hello", "en-US", None, '"hello"@en-US'),
        ("café", "fr", None, '"café"@fr'),
        (
            "123",
            None,
            "http://www.w3.org/2001/XMLSchema#integer",
            '"123"^^<http://www.w3.org/2001/XMLSchema#integer>',
        ),
        (
            "3.14",
            None,
            "http://www.w3.org/2001/XMLSchema#float",
            '"3.14"^^<http://www.w3.org/2001/XMLSchema#float>',
        ),
        (
            "true",
            None,
            "http://www.w3.org/2001/XMLSchema#boolean",
            '"true"^^<http://www.w3.org/2001/XMLSchema#boolean>',
        ),
        (
            "2023-01-01",
            None,
            "http://www.w3.org/2001/XMLSchema#date",
            '"2023-01-01"^^<http://www.w3.org/2001/XMLSchema#date>',
        ),
        (
            "hello",
            None,
            "http://www.w3.org/2001/XMLSchema#string",
            '"hello"^^<http://www.w3.org/2001/XMLSchema#string>',
        ),
        ('He said "hello"', None, None, '"He said "hello""'),
        ("line1\nline2", None, None, '"line1\nline2"'),
        ("tab\there", None, None, '"tab\there"'),
        ("backslash\\here", None, None, '"backslash\\here"'),
        ("résumé", "fr", None, '"résumé"@fr'),
        ("München", "de", None, '"München"@de'),
        ("日本語", "ja", None, '"日本語"@ja'),
        ("🚀", None, None, '"🚀"'),
        (" ", None, None, '" "'),
        ("  whitespace  ", None, None, '"  whitespace  "'),
        ("", "en", None, '""@en'),
        (
            "",
            None,
            "http://www.w3.org/2001/XMLSchema#string",
            '""^^<http://www.w3.org/2001/XMLSchema#string>',
        ),
    ],
)
def test_literal_str(lex: str, langtag: str, datatype: str, expected_str: str) -> None:
    """Test string representation of literals."""
    literal = Literal(lex, langtag=langtag, datatype=datatype)
    assert str(literal) == expected_str


@pytest.mark.parametrize(
    (
        "lex",
        "langtag",
        "datatype",
        "expected_repr",
    ),
    [
        ("hello", None, None, "Literal('hello', langtag=None, datatype=None)"),
        ("", None, None, "Literal('', langtag=None, datatype=None)"),
        ("123", None, None, "Literal('123', langtag=None, datatype=None)"),
        ("hello", "en", None, "Literal('hello', langtag='en', datatype=None)"),
        ("bonjour", "fr", None, "Literal('bonjour', langtag='fr', datatype=None)"),
        ("café", "fr", None, "Literal('café', langtag='fr', datatype=None)"),
        ("hello", "en-US", None, "Literal('hello', langtag='en-US', datatype=None)"),
        (
            "123",
            None,
            "http://www.w3.org/2001/XMLSchema#integer",
            "Literal('123', langtag=None, datatype='http://www.w3.org/2001/XMLSchema#integer')",
        ),
        (
            "3.14",
            None,
            "http://www.w3.org/2001/XMLSchema#float",
            "Literal('3.14', langtag=None, datatype='http://www.w3.org/2001/XMLSchema#float')",
        ),
        (
            "true",
            None,
            "http://www.w3.org/2001/XMLSchema#boolean",
            "Literal('true', langtag=None, datatype='http://www.w3.org/2001/XMLSchema#boolean')",
        ),
        (
            'He said "hello"',
            None,
            None,
            "Literal('He said \"hello\"', langtag=None, datatype=None)",
        ),
        (
            "line1\nline2",
            None,
            None,
            "Literal('line1\\nline2', langtag=None, datatype=None)",
        ),
        ("tab\there", None, None, "Literal('tab\\there', langtag=None, datatype=None)"),
        ("résumé", "fr", None, "Literal('résumé', langtag='fr', datatype=None)"),
        ("日本語", "ja", None, "Literal('日本語', langtag='ja', datatype=None)"),
        ("🚀", None, None, "Literal('🚀', langtag=None, datatype=None)"),
        (" ", None, None, "Literal(' ', langtag=None, datatype=None)"),
        ("", "en", None, "Literal('', langtag='en', datatype=None)"),
        (
            "",
            None,
            "http://www.w3.org/2001/XMLSchema#string",
            "Literal('', langtag=None, datatype='http://www.w3.org/2001/XMLSchema#string')",
        ),
    ],
)
def test_literal_repr(
    lex: str, langtag: str, datatype: str, expected_repr: str
) -> None:
    literal = Literal(lex, langtag=langtag, datatype=datatype)
    assert repr(literal) == expected_repr


@pytest.mark.parametrize(
    (
        "iri",
        "expected_str",
    ),
    [
        ("http://example.com", "<http://example.com>"),
        ("https://example.com", "<https://example.com>"),
        ("http://example.com/resource", "<http://example.com/resource>"),
        (
            "https://example.com/path/to/resource",
            "<https://example.com/path/to/resource>",
        ),
        ("http://example.com#fragment", "<http://example.com#fragment>"),
        ("http://example.com?query=value", "<http://example.com?query=value>"),
        (
            "http://example.com/resource?query=value#fragment",
            "<http://example.com/resource?query=value#fragment>",
        ),
        ("ftp://example.com", "<ftp://example.com>"),
        ("mailto:user@example.com", "<mailto:user@example.com>"),
        ("file:///path/to/file", "<file:///path/to/file>"),
        ("urn:isbn:0451450523", "<urn:isbn:0451450523>"),
        ("http://example.com/résource", "<http://example.com/résource>"),
        (
            "http://example.com/resource with spaces",
            "<http://example.com/resource with spaces>",
        ),
        ("http://example.com/München", "<http://example.com/München>"),
        ("http://example.com/日本語", "<http://example.com/日本語>"),
        ("", "<>"),
        ("relative/path", "<relative/path>"),
        ("//example.com", "<//example.com>"),
        (":", "<:>"),
        (
            "http://example.com/very/long/path/to/resource/with/many/segments",
            "<http://example.com/very/long/path/to/resource/with/many/segments>",
        ),
        (
            "http://example.com/path-with-dashes",
            "<http://example.com/path-with-dashes>",
        ),
        (
            "http://example.com/path_with_underscores",
            "<http://example.com/path_with_underscores>",
        ),
        ("http://example.com/path.with.dots", "<http://example.com/path.with.dots>"),
        (
            "http://example.com/path~with~tildes",
            "<http://example.com/path~with~tildes>",
        ),
        (
            "http://example.com/path%20with%20encoded%20spaces",
            "<http://example.com/path%20with%20encoded%20spaces>",
        ),
        ("http://example.com/path%C3%A9", "<http://example.com/path%C3%A9>"),
    ],
)
def test_iri_str_representation(iri: str, expected_str: str) -> None:
    iri_obj = IRI(iri)
    assert str(iri_obj) == expected_str


@pytest.mark.parametrize(
    (
        "iri",
        "expected_repr",
    ),
    [
        ("http://example.com", "IRI(http://example.com)"),
        ("https://example.com", "IRI(https://example.com)"),
        ("http://example.com/resource", "IRI(http://example.com/resource)"),
        (
            "https://example.com/path/to/resource",
            "IRI(https://example.com/path/to/resource)",
        ),
        ("http://example.com#fragment", "IRI(http://example.com#fragment)"),
        ("http://example.com?query=value", "IRI(http://example.com?query=value)"),
        (
            "http://example.com/resource?query=value#fragment",
            "IRI(http://example.com/resource?query=value#fragment)",
        ),
        ("ftp://example.com", "IRI(ftp://example.com)"),
        ("mailto:user@example.com", "IRI(mailto:user@example.com)"),
        ("file:///path/to/file", "IRI(file:///path/to/file)"),
        ("urn:isbn:0451450523", "IRI(urn:isbn:0451450523)"),
        ("http://example.com/résource", "IRI(http://example.com/résource)"),
        (
            "http://example.com/resource with spaces",
            "IRI(http://example.com/resource with spaces)",
        ),
        ("http://example.com/München", "IRI(http://example.com/München)"),
        ("http://example.com/日本語", "IRI(http://example.com/日本語)"),
        ("", "IRI()"),
        ("relative/path", "IRI(relative/path)"),
        ("//example.com", "IRI(//example.com)"),
        (":", "IRI(:)"),
        (
            "http://example.com/very/long/path/to/resource/with/many/segments",
            "IRI(http://example.com/very/long/path/to/resource/with/many/segments)",
        ),
        (
            "http://example.com/path-with-dashes",
            "IRI(http://example.com/path-with-dashes)",
        ),
        (
            "http://example.com/path_with_underscores",
            "IRI(http://example.com/path_with_underscores)",
        ),
        ("http://example.com/path.with.dots", "IRI(http://example.com/path.with.dots)"),
        (
            "http://example.com/path~with~tildes",
            "IRI(http://example.com/path~with~tildes)",
        ),
        (
            "http://example.com/path%20with%20encoded%20spaces",
            "IRI(http://example.com/path%20with%20encoded%20spaces)",
        ),
        ("http://example.com/path%C3%A9", "IRI(http://example.com/path%C3%A9)"),
    ],
)
def test_iri_repr(iri: str, expected_repr: str) -> None:
    iri_obj = IRI(iri)
    assert repr(iri_obj) == expected_repr


def test_parse_serialize() -> None:
    sink = GenericStatementSink()
    input_file_path = Path(
        "./tests/integration_tests/test_examples/temp/flat_output.jelly"
    )
    output_file_path = Path(
        "./tests/integration_tests/test_examples/temp/temp_output.jelly"
    )
    with input_file_path.open("rb") as file:
        sink.parse(file)
    with output_file_path.open("wb") as file:
        sink.serialize(file)
    new_sink = GenericStatementSink()
    with output_file_path.open("rb") as file:
        new_sink.parse(file)
    assert len(new_sink) > 0
    assert len(sink) > 0
    assert len(new_sink) == len(sink)
    for s_in, s_out in zip(sink.store, new_sink.store):
        assert repr(s_in) == repr(s_out)



from __future__ import annotations

from collections.abc import Generator, Iterable
from itertools import chain
from typing import IO, Any, Callable, Union
from typing_extensions import override

from pyjelly import jelly
from pyjelly.errors import JellyConformanceError
from pyjelly.integrations.generic.generic_sink import (
    IRI,
    BlankNode,
    GenericStatementSink,
    Literal,
    Prefix,
    Quad,
    Triple,
)
from pyjelly.parse.decode import Adapter, Decoder, ParserOptions
from pyjelly.parse.ioutils import get_options_and_frames

Statement = Union[Triple, Quad]


class GenericStatementSinkAdapter(Adapter):
    """
    Implement Adapter for generic statements.

    Notes:
        Returns custom RDF terms expected by GenericStatementSink,
        handles namespace declarations, and quoted triples.

    Args:
        Adapter (_type_): base Adapter class

    """

    @override
    def iri(self, iri: str) -> IRI:
        return IRI(iri)

    @override
    def bnode(self, bnode: str) -> BlankNode:
        return BlankNode(bnode)

    @override
    def default_graph(self) -> str:
        return ""

    @override
    def literal(
        self,
        lex: str,
        language: str | None = None,
        datatype: str | None = None,
    ) -> Literal:
        return Literal(lex, language, datatype)

    @override
    def namespace_declaration(self, name: str, iri: str) -> Prefix:
        return Prefix(name, self.iri(iri))

    @override
    def quoted_triple(self, terms: Iterable[Any]) -> Triple:
        return Triple(*terms)


class GenericTriplesAdapter(GenericStatementSinkAdapter):
    """
    Triples adapted implementation for GenericStatementSink.

    Args:
        GenericStatementSinkAdapter (_type_): base GenericStatementSink
            adapter implementation that handles terms and namespaces.

    """

    def __init__(
        self,
        options: ParserOptions,
    ) -> None:
        super().__init__(options=options)

    @override
    def triple(self, terms: Iterable[Any]) -> Triple:
        return Triple(*terms)


class GenericQuadsBaseAdapter(GenericStatementSinkAdapter):
    def __init__(self, options: ParserOptions) -> None:
        super().__init__(options=options)


class GenericQuadsAdapter(GenericQuadsBaseAdapter):
    """
    Extends GenericQuadsBaseAdapter for QUADS physical type.

    Args:
        GenericQuadsBaseAdapter (_type_): quads adapter that handles
            base quads processing.

    """

    @override
    def quad(self, terms: Iterable[Any]) -> Quad:
        return Quad(*terms)


class GenericGraphsAdapter(GenericQuadsBaseAdapter):
    """
    Extends GenericQuadsBaseAdapter for GRAPHS physical type.

    Notes:
        introduces graph start/end, checks if graph exists.

    Args:
        GenericQuadsBaseAdapter (_type_): quads adapter that handles
            base quads processing.

    Raises:
        JellyConformanceError: raised if graph start message was not received.

    """

    _graph_id: str | None

    def __init__(
        self,
        options: ParserOptions,
    ) -> None:
        super().__init__(options=options)
        self._graph_id = None

    @property
    def graph(self) -> None:
        if self._graph_id is None:
            msg = "new graph was not started"
            raise JellyConformanceError(msg)

    @override
    def graph_start(self, graph_id: str) -> None:
        self._graph_id = graph_id

    @override
    def triple(self, terms: Iterable[Any]) -> Quad:
        return Quad(*chain(terms, [self._graph_id]))

    @override
    def graph_end(self) -> None:
        self._graph_id = None


def parse_triples_stream(
    frames: Iterable[jelly.RdfStreamFrame],
    options: ParserOptions,
) -> Generator[Iterable[Triple | Prefix]]:
    """
    Parse flat triple stream.

    Args:
        frames (Iterable[jelly.RdfStreamFrame]): iterator over stream frames
        options (ParserOptions): stream options

    Yields:
        Generator[Iterable[Triple | Prefix]]:
            Generator of iterables of Triple or Prefix objects,
            one iterable per frame.

    """
    adapter = GenericTriplesAdapter(options)
    decoder = Decoder(adapter=adapter)
    for frame in frames:
        yield decoder.iter_rows(frame)
    return


def parse_quads_stream(
    frames: Iterable[jelly.RdfStreamFrame],
    options: ParserOptions,
) -> Generator[Iterable[Quad | Prefix]]:
    """
    Parse flat quads stream.

    Args:
        frames (Iterable[jelly.RdfStreamFrame]): iterator over stream frames
        options (ParserOptions): stream options

    Yields:
        Generator[Iterable[Quad | Prefix]]:
            Generator of iterables of Quad or Prefix objects,
            one iterable per frame.

    """
    adapter_class: type[GenericQuadsBaseAdapter]
    if options.stream_types.physical_type == jelly.PHYSICAL_STREAM_TYPE_QUADS:
        adapter_class = GenericQuadsAdapter
    else:
        adapter_class = GenericGraphsAdapter
    adapter = adapter_class(options=options)
    decoder = Decoder(adapter=adapter)
    for frame in frames:
        yield decoder.iter_rows(frame)
    return


def parse_jelly_grouped(
    inp: IO[bytes],
    sink_factory: Callable[[], GenericStatementSink] = lambda: GenericStatementSink(),
) -> Generator[GenericStatementSink]:
    """
    Take a jelly file and return generators of generic statements sinks.

    Yields one generic statements sink per frame.

    Args:
        inp (IO[bytes]): input jelly buffered binary stream
        sink_factory (Callable): lambda to construct a statement sink.
            By default creates an empty in-memory GenericStatementSink.

    Raises:
        NotImplementedError: is raised if a physical type is not implemented

    Yields:
        Generator[GenericStatementSink]:
            returns generators for GenericStatementSink, regardless of stream type.

    """
    options, frames = get_options_and_frames(inp)
    if options.stream_types.physical_type == jelly.PHYSICAL_STREAM_TYPE_TRIPLES:
        for graph in parse_triples_stream(
            frames=frames,
            options=options,
        ):
            sink = sink_factory()
            for graph_item in graph:
                if isinstance(graph_item, Prefix):
                    sink.bind(graph_item.prefix, graph_item.iri)
                else:
                    sink.add(graph_item)
            yield sink
        return
    elif options.stream_types.physical_type in (
        jelly.PHYSICAL_STREAM_TYPE_QUADS,
        jelly.PHYSICAL_STREAM_TYPE_GRAPHS,
    ):
        for dataset in parse_quads_stream(
            frames=frames,
            options=options,
        ):
            sink = sink_factory()
            for item in dataset:
                if isinstance(item, Prefix):
                    sink.bind(item.prefix, item.iri)
                else:
                    sink.add(item)
            yield sink
        return

    physical_type_name = jelly.PhysicalStreamType.Name(
        options.stream_types.physical_type
    )
    msg = f"the stream type {physical_type_name} is not supported "
    raise NotImplementedError(msg)


def parse_jelly_to_graph(
    inp: IO[bytes],
    sink_factory: Callable[[], GenericStatementSink] = lambda: GenericStatementSink(),
) -> GenericStatementSink:
    """
    Add statements from Generator to GenericStatementSink.

    Args:
        inp (IO[bytes]): input jelly stream.
        sink_factory (Callable[[], GenericStatementSink]): factory to create
            statement sink.
            By default creates an empty in-memory GenericStatementSink.
            Has no division for datasets/graphs,
            utilizes the same underlying data structures.

    Returns:
        GenericStatementSink: GenericStatementSink with statements.

    """
    options, frames = get_options_and_frames(inp)
    sink = sink_factory()

    for item in parse_jelly_flat(inp=inp, frames=frames, options=options):
        if isinstance(item, Prefix):
            sink.bind(item.prefix, item.iri)
        else:
            sink.add(item)
    return sink


def parse_jelly_flat(
    inp: IO[bytes],
    frames: Iterable[jelly.RdfStreamFrame] | None = None,
    options: ParserOptions | None = None,
) -> Generator[Statement | Prefix]:
    """
    Parse jelly file with FLAT logical type into a Generator of stream events.

    Args:
        inp (IO[bytes]): input jelly buffered binary stream.
        frames (Iterable[jelly.RdfStreamFrame | None):
            jelly frames if read before.
        options (ParserOptions | None): stream options
            if read before.

    Raises:
        NotImplementedError: if physical type is not supported

    Yields:
        Generator[Statement | Prefix]: Generator of stream events

    """
    if not frames or not options:
        options, frames = get_options_and_frames(inp)

    if options.stream_types.physical_type == jelly.PHYSICAL_STREAM_TYPE_TRIPLES:
        for triples in parse_triples_stream(frames=frames, options=options):
            yield from triples
        return
    if options.stream_types.physical_type in (
        jelly.PHYSICAL_STREAM_TYPE_QUADS,
        jelly.PHYSICAL_STREAM_TYPE_GRAPHS,
    ):
        for quads in parse_quads_stream(
            frames=frames,
            options=options,
        ):
            yield from quads
        return
    physical_type_name = jelly.PhysicalStreamType.Name(
        options.stream_types.physical_type
    )
    msg = f"the stream type {physical_type_name} is not supported "
    raise NotImplementedError(msg)



import pytest
from inline_snapshot import snapshot
from pytest_subtests import SubTests

from pyjelly import jelly
from pyjelly.errors import JellyConformanceError
from pyjelly.options import LookupPreset
from pyjelly.serialize.encode import (
    Slot,
    TermEncoder,
    encode_namespace_declaration,
)


def test_encode_literal_fails_with_disabled_datatype_lookup() -> None:
    encoder = TermEncoder(
        lookup_preset=LookupPreset(
            max_names=8,
            max_prefixes=8,
            max_datatypes=0,
        )
    )
    with pytest.raises(
        JellyConformanceError,
        match="datatype lookup cannot be used if disabled",
    ):
        encoder.encode_literal(
            lex="42",
            datatype="http://www.w3.org/2001/XMLSchema#integer",
        )


def test_encode_any_raises_not_implemented() -> None:
    encoder = TermEncoder()
    with pytest.raises(NotImplementedError) as exc:
        encoder.encode_any(123, Slot.subject)
    assert "unsupported term type: <class 'int'>" in str(exc.value)


def test_encode_literal_ok_with_string_and_langtag(subtests: SubTests) -> None:
    encoder = TermEncoder(
        lookup_preset=LookupPreset(
            max_names=8,
            max_prefixes=8,
            max_datatypes=0,
        )
    )

    with subtests.test("xsd:string is skipped by datatype lookup → no error"):
        _, literal = encoder.encode_literal(
            lex="foo",
            datatype="http://www.w3.org/2001/XMLSchema#string",
        )
        assert literal.lex == snapshot("foo")
        assert literal.datatype == snapshot(0)
        assert literal.langtag == snapshot("")

    with subtests.test("no datatype or langtag"):
        _, literal = encoder.encode_literal(lex="bar")
        assert literal.lex == snapshot("bar")
        assert literal.datatype == snapshot(0)
        assert literal.langtag == snapshot("")

    with subtests.test("no datatype or langtag"):
        _, literal = encoder.encode_literal(lex="baz", language="en")
        assert literal.lex == snapshot("baz")
        assert literal.langtag == snapshot("en")
        assert literal.datatype == snapshot(0)


def test_encode_namespace_declaration() -> None:
    encoder = TermEncoder()
    rows = encode_namespace_declaration("ex", "http://example.org/A", encoder)

    assert isinstance(rows[-1].namespace, jelly.RdfNamespaceDeclaration)

    assert any(r.prefix for r in rows[:-1])
    assert any(r.name for r in rows[:-1])



import pytest
from hypothesis import given
from hypothesis import strategies as st

from pyjelly.errors import JellyAssertionError, JellyConformanceError
from pyjelly.options import MAX_LOOKUP_SIZE
from pyjelly.parse.lookup import LookupDecoder


@given(st.integers(min_value=1, max_value=MAX_LOOKUP_SIZE))
def test_lookup_size_ok(size: int) -> None:
    LookupDecoder(lookup_size=size)


@given(st.integers(min_value=MAX_LOOKUP_SIZE + 1))
def test_max_lookup_size_exceeded(size: int) -> None:
    with pytest.raises(JellyAssertionError) as excinfo:
        LookupDecoder(lookup_size=size)
    assert str(excinfo.value) == f"lookup size must be less than {MAX_LOOKUP_SIZE}"


def test_decode_zero_error() -> None:
    dec = LookupDecoder(lookup_size=1)
    dec.last_reused_index = -1
    with pytest.raises(JellyConformanceError):
        dec.decode_name_term_index(0)


def test_datatype_index_zero_error() -> None:
    decoder = LookupDecoder(lookup_size=0)
    with pytest.raises(JellyConformanceError) as excinfo:
        decoder.decode_datatype_term_index(0)
    assert str(excinfo.value) == "0 is not a valid datatype term index"


def test_at_invalid_index() -> None:
    decoder = LookupDecoder(lookup_size=4)
    with pytest.raises(IndexError) as excinfo:
        decoder.at(2)
    assert "invalid resolved index 2" in str(excinfo.value)
    assert decoder.last_reused_index == 2



from __future__ import annotations

import mimetypes
from contextlib import suppress
from dataclasses import dataclass
from typing import Final
from typing_extensions import Self

from pyjelly import jelly
from pyjelly.errors import (
    JellyAssertionError,
    JellyConformanceError,
)

MIN_NAME_LOOKUP_SIZE: Final[int] = 8

MAX_LOOKUP_SIZE: Final[int] = 4096
MIN_VERSION: Final[int] = 1
MAX_VERSION: Final[int] = 2

DEFAULT_NAME_LOOKUP_SIZE: Final[int] = 4000
DEFAULT_PREFIX_LOOKUP_SIZE: Final[int] = 150
DEFAULT_DATATYPE_LOOKUP_SIZE: Final[int] = 32

STRING_DATATYPE_IRI = "http://www.w3.org/2001/XMLSchema#string"

INTEGRATION_SIDE_EFFECTS: bool = True
"""
Whether to allow integration module imports to trigger side effects.

These side effects are cheap and may include populating some registries
for guessing the defaults for external integrations that work with Jelly.
"""

MIMETYPES = ("application/x-jelly-rdf",)


def register_mimetypes(extension: str = ".jelly") -> None:
    """
    Associate files that have Jelly extension with Jelly MIME types.

    >>> register_mimetypes()
    >>> mimetypes.guess_type("out.jelly")
    ('application/x-jelly-rdf', None)
    """
    for mimetype in MIMETYPES:
        mimetypes.add_type(mimetype, extension)


@dataclass(frozen=True)
class LookupPreset:
    max_names: int = DEFAULT_NAME_LOOKUP_SIZE
    max_prefixes: int = DEFAULT_PREFIX_LOOKUP_SIZE
    max_datatypes: int = DEFAULT_DATATYPE_LOOKUP_SIZE

    def __post_init__(self) -> None:
        if self.max_names < MIN_NAME_LOOKUP_SIZE:
            msg = "name lookup size must be at least 8"
            raise JellyConformanceError(msg)

    @classmethod
    def small(cls) -> Self:
        return cls(max_names=128, max_prefixes=32, max_datatypes=32)


@dataclass(frozen=True)
class StreamTypes:
    physical_type: jelly.PhysicalStreamType = jelly.PHYSICAL_STREAM_TYPE_UNSPECIFIED
    logical_type: jelly.LogicalStreamType = jelly.LOGICAL_STREAM_TYPE_UNSPECIFIED

    @property
    def flat(self) -> bool:
        return self.logical_type in (
            jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES,
            jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
        )

    def __repr__(self) -> str:
        """
        Return the representation of StreamTypes.

        >>> repr(StreamTypes(9999, 8888))
        'StreamTypes(9999, 8888)'
        """
        with suppress(ValueError):
            physical_type_name = jelly.PhysicalStreamType.Name(self.physical_type)
            logical_type_name = jelly.LogicalStreamType.Name(self.logical_type)
            return f"StreamTypes({physical_type_name}, {logical_type_name})"
        return f"StreamTypes({self.physical_type}, {self.logical_type})"

    def __post_init__(self) -> None:
        validate_type_compatibility(
            physical_type=self.physical_type,
            logical_type=self.logical_type,
        )


@dataclass(frozen=True)
class StreamParameters:
    generalized_statements: bool = False
    rdf_star: bool = False
    version: int = MAX_VERSION
    delimited: bool = True
    namespace_declarations: bool = False
    stream_name: str = ""

    def __post_init__(self) -> None:
        selected = MAX_VERSION if self.namespace_declarations else MIN_VERSION
        if not (MIN_VERSION <= selected <= MAX_VERSION):
            msg = f"""Error occured while settin up the Stream options.
            Version must be between {MIN_VERSION} and {MAX_VERSION}."""
            raise JellyConformanceError(msg)
        object.__setattr__(self, "version", selected)


TRIPLES_ONLY_LOGICAL_TYPES = {
    jelly.LOGICAL_STREAM_TYPE_GRAPHS,
    jelly.LOGICAL_STREAM_TYPE_SUBJECT_GRAPHS,
    jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES,
}


def validate_type_compatibility(
    physical_type: jelly.PhysicalStreamType,
    logical_type: jelly.LogicalStreamType,
) -> None:
    if (
        physical_type == jelly.PHYSICAL_STREAM_TYPE_UNSPECIFIED
        or logical_type == jelly.LOGICAL_STREAM_TYPE_UNSPECIFIED
    ):
        return
    triples_physical_type = physical_type == jelly.PHYSICAL_STREAM_TYPE_TRIPLES
    triples_logical_type = logical_type in TRIPLES_ONLY_LOGICAL_TYPES
    if triples_physical_type != triples_logical_type:
        physical_type_name = jelly.PhysicalStreamType.Name(physical_type)
        logical_type_name = jelly.LogicalStreamType.Name(logical_type)
        msg = f"{physical_type_name} is not compatible with {logical_type_name}"
        raise JellyAssertionError(msg)



from pyjelly.integrations.generic.generic_sink import (
    IRI,
    BlankNode,
    GenericStatementSink,
    Literal,
    Triple,
)

# Create a generic sink object, identifier is optional
generic_sink = GenericStatementSink(identifier="my_graph")

# Let's add triples one by one
generic_sink.add(
    Triple(
        IRI("http://example.com/subject"),
        IRI("http://example.com/predicate"),
        Literal("Hello", langtag="en"),
    )
)
generic_sink.add(
    Triple(
        BlankNode("B1"),
        IRI("http://example.com/hasName"),
        Literal("Bob"),
    )
)

# Write into a Jelly file
with open("output.jelly", "wb") as out_file:
    generic_sink.serialize(out_file)

print("All done.")



from __future__ import annotations

from collections import deque
from collections.abc import Generator
from typing import IO, NamedTuple, Union

DEFAULT_GRAPH_IDENTIFIER = ""


class BlankNode:
    """Class for blank nodes, storing BN's identifier as a string."""

    def __init__(self, identifier: str) -> None:
        self._identifier: str = identifier

    def __str__(self) -> str:
        return f"_:{self._identifier}"

    def __repr__(self) -> str:
        return f"BlankNode(identifier={self._identifier})"


class IRI:
    """Class for IRIs, storing IRI as a string."""

    def __init__(self, iri: str) -> None:
        self._iri: str = iri

    def __str__(self) -> str:
        return f"<{self._iri}>"

    def __repr__(self) -> str:
        return f"IRI({self._iri})"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, IRI):
            return self._iri == other._iri
        return False


class Literal:
    """
    Class for literals.

    Notes:
        Consists of: lexical form, and optional language tag and datatype.
        All parts of literal are stored as strings.

    """

    def __init__(
        self, lex: str, langtag: str | None = None, datatype: str | None = None
    ) -> None:
        self._lex: str = lex
        self._langtag: str | None = langtag
        self._datatype: str | None = datatype

    def __str__(self) -> str:
        suffix = ""
        if self._langtag:
            suffix = f"@{self._langtag}"
        elif self._datatype:
            suffix = f"^^<{self._datatype}>"
        return f'"{self._lex}"{suffix}'

    def __repr__(self) -> str:
        return (
            f"Literal({self._lex!r}, langtag={self._langtag!r}, "
            f"datatype={self._datatype!r})"
        )


Node = Union[BlankNode, IRI, Literal, "Triple", str]


TRIPLE_ARITY = 3


class Triple(NamedTuple):
    """Class for RDF triples."""

    s: Node
    p: Node
    o: Node


class Quad(NamedTuple):
    """Class for RDF quads."""

    s: Node
    p: Node
    o: Node
    g: Node


class Prefix(NamedTuple):
    """Class for generic namespace declaration."""

    prefix: str
    iri: IRI


class GenericStatementSink:
    _store: deque[Triple | Quad]

    def __init__(self, identifier: Node = DEFAULT_GRAPH_IDENTIFIER) -> None:
        """
        Initialize statements storage, namespaces dictionary, and parser.

        Notes:
            _store preserves the order of statements.

        Args:
            identifier (str, optional): Identifier for a sink.
                Defaults to DEFAULT_GRAPH_IDENTIFIER.

        """
        self._store: deque[Triple | Quad] = deque()
        self._namespaces: dict[str, IRI] = {}
        self._identifier = identifier

    def add(self, statement: Triple | Quad) -> None:
        self._store.append(statement)

    def bind(self, prefix: str, namespace: IRI) -> None:
        self._namespaces.update({prefix: namespace})

    def __iter__(self) -> Generator[Triple | Quad]:
        yield from self._store

    def __len__(self) -> int:
        return len(self._store)

    @property
    def namespaces(self) -> Generator[tuple[str, IRI]]:
        yield from self._namespaces.items()

    @property
    def identifier(self) -> Node:
        return self._identifier

    @property
    def store(self) -> Generator[Triple | Quad]:
        yield from self._store

    @property
    def is_triples_sink(self) -> bool:
        """
        Check if the sink contains triples or quads.

        Returns:
            bool: true, if length of statement is 3.

        """
        return len(self._store[0]) == TRIPLE_ARITY

    def parse(self, input_file: IO[bytes]) -> None:
        from pyjelly.integrations.generic.parse import parse_jelly_to_graph

        parsed_result = parse_jelly_to_graph(input_file)
        self._store = parsed_result._store
        self._namespaces = parsed_result._namespaces
        self._identifier = parsed_result._identifier

    def serialize(self, output_file: IO[bytes]) -> None:
        from pyjelly.integrations.generic.serialize import grouped_stream_to_file

        grouped_stream_to_file((sink for sink in [self]), output_file)



# ruff: noqa: D205,D400,D415
from __future__ import annotations

import rdflib


def fixup_term(term: rdflib.Node) -> None:
    """
    >>> l1 = rdflib.Literal("foo", datatype=rdflib.XSD.string)
    >>> l2 = rdflib.Literal("foo")
    >>> l1 == l2
    False
    >>> fixup_term(l1)
    >>> fixup_term(l2)
    >>> l1 == l2
    True
    """
    # workaround #99
    if isinstance(term, rdflib.Literal) and term.datatype == rdflib.XSD.string:
        term._datatype = None


def fixup_graph(graph: rdflib.Graph) -> None:
    for stmt in graph:
        for term in stmt:
            fixup_term(term)



from __future__ import annotations

from collections.abc import Generator, Iterable
from itertools import chain
from typing import IO, Any, Callable, Union
from typing_extensions import Never, Self, override

import rdflib
from rdflib import Node
from rdflib.graph import DATASET_DEFAULT_GRAPH_ID, Dataset, Graph
from rdflib.parser import InputSource
from rdflib.parser import Parser as RDFLibParser

from pyjelly import jelly
from pyjelly.errors import JellyConformanceError
from pyjelly.options import StreamTypes
from pyjelly.parse.decode import Adapter, Decoder, ParserOptions
from pyjelly.parse.ioutils import get_options_and_frames

GraphName = Union[rdflib.URIRef, rdflib.BNode, str]


class Triple(tuple[Node, Node, Node]):
    """
    Describe RDFLib triple.

    Args:
        tuple (Node, Node, Node): s/p/o tuple of RDFLib Nodes.

    Returns:
        Triple: triple as tuple.

    """

    __slots__ = ()

    def __new__(cls, s: Node, p: Node, o: Node) -> Self:
        return tuple.__new__(cls, (s, p, o))

    @property
    def s(self) -> Node:
        return self[0]

    @property
    def p(self) -> Node:
        return self[1]

    @property
    def o(self) -> Node:
        return self[2]


class Quad(tuple[Node, Node, Node, GraphName]):
    """
    Describe RDFLib quad.

    Args:
        tuple (Node, Node, Node, GraphName):
            s/p/o/g as a tuple of RDFLib nodes and a GraphName,

    Returns:
        Quad: quad as tuple.

    """

    __slots__ = ()

    def __new__(cls, s: Node, p: Node, o: Node, g: GraphName) -> Self:
        return tuple.__new__(cls, (s, p, o, g))

    @property
    def s(self) -> Node:
        return self[0]

    @property
    def p(self) -> Node:
        return self[1]

    @property
    def o(self) -> Node:
        return self[2]

    @property
    def g(self) -> GraphName:
        return self[3]


Statement = Union[Triple, Quad]


class Prefix(tuple[str, rdflib.URIRef]):
    """
    Describe RDF Prefix(i.e, namespace declaration).

    Args:
        tuple (str, rdflib.URIRef): expects prefix as a string,
            and full namespace URI as Rdflib.URIRef.

    Returns:
        Prefix: prefix as tuple(prefix, iri).

    """

    __slots__ = ()

    def __new__(cls, prefix: str, iri: rdflib.URIRef) -> Self:
        return tuple.__new__(cls, (prefix, iri))

    @property
    def prefix(self) -> str:
        return self[0]

    @property
    def iri(self) -> rdflib.URIRef:
        return self[1]


class RDFLibAdapter(Adapter):
    """
    RDFLib adapter class, is extended by triples and quads implementations.

    Args:
        Adapter (): abstract adapter class

    """

    @override
    def iri(self, iri: str) -> rdflib.URIRef:
        return rdflib.URIRef(iri)

    @override
    def bnode(self, bnode: str) -> rdflib.BNode:
        return rdflib.BNode(bnode)

    @override
    def default_graph(self) -> rdflib.URIRef:
        return DATASET_DEFAULT_GRAPH_ID

    @override
    def literal(
        self,
        lex: str,
        language: str | None = None,
        datatype: str | None = None,
    ) -> rdflib.Literal:
        return rdflib.Literal(lex, lang=language, datatype=datatype)

    @override
    def namespace_declaration(self, name: str, iri: str) -> Prefix:
        return Prefix(name, self.iri(iri))


def _adapter_missing(feature: str, *, stream_types: StreamTypes) -> Never:
    """
    Raise error if functionality is missing in adapter.

    Args:
        feature (str): function which is not implemented
        stream_types (StreamTypes): what combination of physical/logical types
            triggered the error

    Raises:
        NotImplementedError: raises error with message with missing functionality
            and types encountered

    Returns:
        Never: only raises errors

    """
    physical_type_name = jelly.PhysicalStreamType.Name(stream_types.physical_type)
    logical_type_name = jelly.LogicalStreamType.Name(stream_types.logical_type)
    msg = (
        f"adapter with {physical_type_name} and {logical_type_name} "
        f"does not implement {feature}"
    )
    raise NotImplementedError(msg)


class RDFLibTriplesAdapter(RDFLibAdapter):
    """
    Triples adapter RDFLib implementation.

    Notes: returns triple/namespace declaration as soon as receives them.
    """

    def __init__(
        self,
        options: ParserOptions,
    ) -> None:
        super().__init__(options=options)

    @override
    def triple(self, terms: Iterable[Any]) -> Triple:
        return Triple(*terms)


class RDFLibQuadsBaseAdapter(RDFLibAdapter):
    def __init__(self, options: ParserOptions) -> None:
        super().__init__(options=options)


class RDFLibQuadsAdapter(RDFLibQuadsBaseAdapter):
    """
    Extended RDFLib adapter for the QUADS physical type.

    Args:
        RDFLibQuadsBaseAdapter (RDFLibAdapter): base quads adapter
            (shared with graphs physical type)

    """

    @override
    def quad(self, terms: Iterable[Any]) -> Quad:
        return Quad(*terms)


class RDFLibGraphsAdapter(RDFLibQuadsBaseAdapter):
    """
    Extension of RDFLibQuadsBaseAdapter for the GRAPHS physical type.

    Notes: introduces graph start/end, checks if graph exists.

    Args:
        RDFLibQuadsBaseAdapter (RDFLibAdapter): base adapter for quads management.

    Raises:
        JellyConformanceError: if no graph_start was encountered

    """

    _graph_id: str | None

    def __init__(
        self,
        options: ParserOptions,
    ) -> None:
        super().__init__(options=options)
        self._graph_id = None

    @property
    def graph(self) -> None:
        if self._graph_id is None:
            msg = "new graph was not started"
            raise JellyConformanceError(msg)

    @override
    def graph_start(self, graph_id: str) -> None:
        self._graph_id = graph_id

    @override
    def triple(self, terms: Iterable[Any]) -> Quad:
        return Quad(*chain(terms, [self._graph_id]))

    @override
    def graph_end(self) -> None:
        self._graph_id = None


def parse_triples_stream(
    frames: Iterable[jelly.RdfStreamFrame],
    options: ParserOptions,
) -> Generator[Iterable[Triple | Prefix]]:
    """
    Parse flat triple stream.

    Args:
        frames (Iterable[jelly.RdfStreamFrame]): iterator over stream frames
        options (ParserOptions): stream options

    Yields:
        Generator[Iterable[Triple | Prefix]]:
            Generator of iterables of Triple or Prefix objects,
            one iterable per frame.

    """
    adapter = RDFLibTriplesAdapter(options)
    decoder = Decoder(adapter=adapter)
    for frame in frames:
        yield decoder.iter_rows(frame)
    return


def parse_quads_stream(
    frames: Iterable[jelly.RdfStreamFrame],
    options: ParserOptions,
) -> Generator[Iterable[Quad | Prefix]]:
    """
    Parse flat quads stream.

    Args:
        frames (Iterable[jelly.RdfStreamFrame]): iterator over stream frames
        options (ParserOptions): stream options

    Yields:
        Generator[Iterable[Quad | Prefix]]:
            Generator of iterables of Quad or Prefix objects,
            one iterable per frame.

    """
    adapter_class: type[RDFLibQuadsBaseAdapter]
    if options.stream_types.physical_type == jelly.PHYSICAL_STREAM_TYPE_QUADS:
        adapter_class = RDFLibQuadsAdapter
    else:
        adapter_class = RDFLibGraphsAdapter
    adapter = adapter_class(options=options)
    decoder = Decoder(adapter=adapter)
    for frame in frames:
        yield decoder.iter_rows(frame)
    return


def parse_jelly_grouped(
    inp: IO[bytes],
    graph_factory: Callable[[], Graph] = lambda: Graph(),
    dataset_factory: Callable[[], Dataset] = lambda: Dataset(),
) -> Generator[Graph] | Generator[Dataset]:
    """
    Take jelly file and return generators based on the detected physical type.

    Yields one graph/dataset per frame.

    Args:
        inp (IO[bytes]): input jelly buffered binary stream
        graph_factory (Callable): lambda to construct a Graph.
            By default creates an empty in-memory Graph,
            but you can pass something else here.
        dataset_factory (Callable): lambda to construct a Dataset.
            By default creates an empty in-memory Dataset,
            but you can pass something else here.

    Raises:
        NotImplementedError: is raised if a physical type is not implemented

    Yields:
        Generator[Graph] | Generator[Dataset]:
            returns generators for graphs/datasets based on the type of input

    """
    options, frames = get_options_and_frames(inp)
    if options.stream_types.physical_type == jelly.PHYSICAL_STREAM_TYPE_TRIPLES:
        for graph in parse_triples_stream(
            frames=frames,
            options=options,
        ):
            sink = graph_factory()
            for graph_item in graph:
                if isinstance(graph_item, Prefix):
                    sink.bind(graph_item.prefix, graph_item.iri)
                else:
                    sink.add(graph_item)
            yield sink
        return
    elif options.stream_types.physical_type in (
        jelly.PHYSICAL_STREAM_TYPE_QUADS,
        jelly.PHYSICAL_STREAM_TYPE_GRAPHS,
    ):
        for dataset in parse_quads_stream(
            frames=frames,
            options=options,
        ):
            sink = dataset_factory()
            for item in dataset:
                if isinstance(item, Prefix):
                    sink.bind(item.prefix, item.iri)
                else:
                    s, p, o, graph_name = item
                    context = sink.get_context(graph_name)
                    sink.add((s, p, o, context))
            yield sink
        return

    physical_type_name = jelly.PhysicalStreamType.Name(
        options.stream_types.physical_type
    )
    msg = f"the stream type {physical_type_name} is not supported "
    raise NotImplementedError(msg)


def parse_jelly_to_graph(
    inp: IO[bytes],
    graph_factory: Callable[[], Graph] = lambda: Graph(),
    dataset_factory: Callable[[], Dataset] = lambda: Dataset(),
) -> Graph | Dataset:
    """
    Add statements from Generator to provided Graph/Dataset.

    Args:
        inp (IO[bytes]): input jelly stream.
        graph_factory (Callable[[], Graph]): factory to create Graph.
            By default creates an empty in-memory Graph,
            but you can pass something else here.
        dataset_factory (Callable[[], Dataset]): factory to create Dataset.
            By default creates an empty in-memory Dataset,
            but you can pass something else here.

    Returns:
        Dataset | Graph: Dataset or Graph with statements.

    """
    options, frames = get_options_and_frames(inp)

    if options.stream_types.physical_type == jelly.PHYSICAL_STREAM_TYPE_TRIPLES:
        sink = graph_factory()
    if options.stream_types.physical_type in (
        jelly.PHYSICAL_STREAM_TYPE_QUADS,
        jelly.PHYSICAL_STREAM_TYPE_GRAPHS,
    ):
        quad_sink = dataset_factory()
        sink = quad_sink

    for item in parse_jelly_flat(inp=inp, frames=frames, options=options):
        if isinstance(item, Prefix):
            sink.bind(item.prefix, item.iri)
        if isinstance(item, Triple):
            sink.add(item)
        if isinstance(item, Quad):
            s, p, o, graph_name = item
            context = quad_sink.get_context(graph_name)
            quad_sink.add((s, p, o, context))
    return sink


def parse_jelly_flat(
    inp: IO[bytes],
    frames: Iterable[jelly.RdfStreamFrame] | None = None,
    options: ParserOptions | None = None,
) -> Generator[Statement | Prefix]:
    """
    Parse jelly file with FLAT logical type into a Generator of stream events.

    Args:
        inp (IO[bytes]): input jelly buffered binary stream.
        frames (Iterable[jelly.RdfStreamFrame | None):
            jelly frames if read before.
        options (ParserOptions | None): stream options
            if read before.

    Raises:
        NotImplementedError: if physical type is not supported

    Yields:
        Generator[Statement | Prefix]: Generator of stream events

    """
    if not frames or not options:
        options, frames = get_options_and_frames(inp)

    if options.stream_types.physical_type == jelly.PHYSICAL_STREAM_TYPE_TRIPLES:
        for triples in parse_triples_stream(frames=frames, options=options):
            yield from triples
        return
    if options.stream_types.physical_type in (
        jelly.PHYSICAL_STREAM_TYPE_QUADS,
        jelly.PHYSICAL_STREAM_TYPE_GRAPHS,
    ):
        for quads in parse_quads_stream(
            frames=frames,
            options=options,
        ):
            yield from quads
        return
    physical_type_name = jelly.PhysicalStreamType.Name(
        options.stream_types.physical_type
    )
    msg = f"the stream type {physical_type_name} is not supported "
    raise NotImplementedError(msg)


class RDFLibJellyParser(RDFLibParser):
    def parse(self, source: InputSource, sink: Graph) -> None:
        """
        Parse jelly file into provided RDFLib Graph.

        Args:
            source (InputSource): jelly file as buffered binary stream InputSource obj
            sink (Graph): RDFLib Graph

        Raises:
            TypeError: raises error if invalid input

        """
        inp = source.getByteStream()  # type: ignore[no-untyped-call]
        if inp is None:
            msg = "expected source to be a stream of bytes"
            raise TypeError(msg)
        parse_jelly_to_graph(
            inp,
            graph_factory=lambda: Graph(store=sink.store, identifier=sink.identifier),
            dataset_factory=lambda: Dataset(store=sink.store),
        )



import urllib.request
from neo4j import GraphDatabase
from rdflib import Graph
from rdflib_neo4j import Neo4jStoreConfig, Neo4jStore, HANDLE_VOCAB_URI_STRATEGY

# Example file from the Riverbench
example_file, _ = urllib.request.urlretrieve("https://w3id.org/riverbench/v/dev.jelly")

# Please introduce your credentials
aura_db_uri = "aura_db_uri"
aura_db_username = "neo4j"
aura_db_pwd = "aura_db_pwd"

# Prepare the authentication data to the AuraDB
auth_data = {
    "uri": aura_db_uri,
    "database": "neo4j",
    "user": aura_db_username,
    "pwd": aura_db_pwd,
}

# Prepare the configuration for Neo4jStore object
config = Neo4jStoreConfig(
    auth_data=auth_data,
    handle_vocab_uri_strategy=HANDLE_VOCAB_URI_STRATEGY.IGNORE,
    batching=True,
)

# Make a graph with Neo4jStore object
neo4j_sink = Graph(store=Neo4jStore(config=config))

# Parse the file into the AuraDB
neo4j_sink.parse(example_file, format="jelly")

# Close the stream
neo4j_sink.close(True)

# Cypher query to check loaded data
with GraphDatabase.driver(
    aura_db_uri, auth=(aura_db_username, aura_db_pwd)
).session() as session:
    count = session.run("MATCH ()-[r]->() RETURN count(r) AS c").single()["c"]
    print(f"Loaded {count} triples")
    for r in session.run(
        "MATCH (s)-[r]->(o) RETURN s.uri AS s, type(r) AS p, coalesce(o.uri,o.name) AS o LIMIT 5"
    ):
        print(r["s"], f"-[{r['p']}]->", r["o"])

print("All done.")



# ruff: noqa: I001
from __future__ import annotations
from typing import cast
from collections.abc import Generator
from functools import singledispatch
from typing import Any, IO
from typing_extensions import override
from itertools import chain
from pyjelly.integrations.rdflib.parse import Quad, Triple

import rdflib
from rdflib import Graph
from rdflib.graph import DATASET_DEFAULT_GRAPH_ID, Dataset, QuotedGraph
from rdflib.serializer import Serializer as RDFLibSerializer

from pyjelly import jelly
from pyjelly.serialize.encode import RowsAndTerm, Slot, TermEncoder
from pyjelly.serialize.ioutils import write_delimited, write_single
from pyjelly.serialize.streams import (
    GraphStream,
    QuadStream,
    SerializerOptions,
    Stream,
    TripleStream,
)  # ruff: enable

QUAD_ARITY = 4


class RDFLibTermEncoder(TermEncoder):
    def encode_any(self, term: object, slot: Slot) -> RowsAndTerm:
        """
        Encode term based on its RDFLib object.

        Args:
            term (object): term to encode
            slot (Slot): its place in statement.

        Returns:
            RowsAndTerm: encoded extra rows and a jelly term to encode

        """
        if slot is Slot.graph and term == DATASET_DEFAULT_GRAPH_ID:
            return self.encode_default_graph()

        if isinstance(term, rdflib.URIRef):
            return self.encode_iri(term)

        if isinstance(term, rdflib.Literal):
            return self.encode_literal(
                lex=str(term),
                language=term.language,
                # `datatype` is cast to `str` explicitly because
                # `URIRef.__eq__` overrides `str.__eq__` in an incompatible manner
                datatype=term.datatype and str(term.datatype),
            )

        if isinstance(term, rdflib.BNode):
            return self.encode_bnode(str(term))

        return super().encode_any(term, slot)  # error if not handled


def namespace_declarations(store: Graph, stream: Stream) -> None:
    for prefix, namespace in store.namespaces():
        stream.namespace_declaration(name=prefix, iri=namespace)


@singledispatch
def stream_frames(
    stream: Stream,
    data: Graph | Generator[Quad | Triple],  # noqa: ARG001
) -> Generator[jelly.RdfStreamFrame]:
    msg = f"invalid stream implementation {stream}"
    raise TypeError(msg)


@stream_frames.register(TripleStream)
def triples_stream_frames(
    stream: TripleStream,
    data: Graph | Dataset | Generator[Triple],
) -> Generator[jelly.RdfStreamFrame]:
    """
    Serialize a Graph/Dataset into jelly frames.

    Args:
        stream (TripleStream): stream that specifies triples processing
        data (Graph | Dataset | Generator[Triple]):
            Graph/Dataset/Statements to serialize.

    Notes:
        if Dataset is given, its graphs are unpacked and iterated over
        if flow is GraphsFrameFlow, emits a frame per graph.

    Yields:
        Generator[jelly.RdfStreamFrame]: jelly frames.

    """
    stream.enroll()
    if isinstance(data, Graph) and stream.options.params.namespace_declarations:
        namespace_declarations(data, stream)

    graphs = (data,) if not isinstance(data, Dataset) else data.graphs()
    for graph in graphs:
        for terms in graph:
            if frame := stream.triple(terms):
                yield frame
        if frame := stream.flow.frame_from_graph():
            yield frame
    if stream.stream_types.flat and (frame := stream.flow.to_stream_frame()):
        yield frame


@stream_frames.register(QuadStream)
def quads_stream_frames(
    stream: QuadStream,
    data: Dataset | Generator[Quad],
) -> Generator[jelly.RdfStreamFrame]:
    """
    Serialize a Dataset into jelly frames.

    Notes:
        Emits one frame per dataset if flow is of DatasetsFrameFlow.

    Args:
        stream (QuadStream): stream that specifies quads processing
        data (Dataset | Generator[Quad]): Dataset to serialize.

    Yields:
        Generator[jelly.RdfStreamFrame]: jelly frames

    """
    stream.enroll()
    if stream.options.params.namespace_declarations:
        namespace_declarations(data, stream)  # type: ignore[arg-type]

    iterator: Generator[Quad, None, None]
    if isinstance(data, Dataset):
        iterator = cast(Generator[Quad, None, None], data.quads())
    else:
        iterator = data

    for terms in iterator:
        if frame := stream.quad(terms):
            yield frame
    if frame := stream.flow.frame_from_dataset():
        yield frame
    if stream.stream_types.flat and (frame := stream.flow.to_stream_frame()):
        yield frame


@stream_frames.register(GraphStream)
def graphs_stream_frames(
    stream: GraphStream,
    data: Dataset | Generator[Quad],
) -> Generator[jelly.RdfStreamFrame]:
    """
    Serialize a Dataset into jelly frames as a stream of graphs.

    Notes:
        If flow of DatasetsFrameFlow type, the whole dataset
        will be encoded into one frame.

    Args:
        stream (GraphStream): stream that specifies graphs processing
        data (Dataset | Generator[Quad]): Dataset to serialize.

    Yields:
        Generator[jelly.RdfStreamFrame]: jelly frames

    """
    stream.enroll()
    if stream.options.params.namespace_declarations:
        namespace_declarations(data, stream)  # type: ignore[arg-type]

    if isinstance(data, Dataset):
        graphs = data.graphs()
    else:
        ds = Dataset()
        for quad in data:
            ctx = ds.get_context(quad.g)
            ctx.add((quad.s, quad.p, quad.o))
        graphs = ds.graphs()

    for graph in graphs:
        yield from stream.graph(graph_id=graph.identifier, graph=graph)

    if frame := stream.flow.frame_from_dataset():
        yield frame
    if stream.stream_types.flat and (frame := stream.flow.to_stream_frame()):
        yield frame


def guess_options(sink: Graph | Dataset) -> SerializerOptions:
    """
    Guess the serializer options based on the store type.

    >>> guess_options(Graph()).logical_type
    1
    >>> guess_options(Dataset()).logical_type
    2
    """
    logical_type = (
        jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS
        if isinstance(sink, Dataset)
        else jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES
    )
    return SerializerOptions(logical_type=logical_type)


def guess_stream(options: SerializerOptions, sink: Graph | Dataset) -> Stream:
    """
    Return an appropriate stream implementation for the given options.

    Notes: if base(!) logical type is GRAPHS and Dataset is given,
        initializes TripleStream

    >>> graph_ser = RDFLibJellySerializer(Graph())
    >>> ds_ser = RDFLibJellySerializer(Dataset())

    >>> type(guess_stream(guess_options(graph_ser.store), graph_ser.store))
    <class 'pyjelly.serialize.streams.TripleStream'>
    >>> type(guess_stream(guess_options(ds_ser.store), ds_ser.store))
    <class 'pyjelly.serialize.streams.QuadStream'>
    """
    stream_cls: type[Stream]
    if (options.logical_type % 10) != jelly.LOGICAL_STREAM_TYPE_GRAPHS and isinstance(
        sink, Dataset
    ):
        stream_cls = QuadStream
    else:
        stream_cls = TripleStream
    return stream_cls.for_rdflib(options=options)


class RDFLibJellySerializer(RDFLibSerializer):
    """
    RDFLib serializer for writing graphs in Jelly RDF stream format.

    Handles streaming RDF terms into Jelly frames using internal encoders.
    Supports only graphs and datasets (not quoted graphs).

    """

    def __init__(self, store: Graph) -> None:
        if isinstance(store, QuotedGraph):
            msg = "N3 format is not supported"
            raise NotImplementedError(msg)
        super().__init__(store)

    @override
    def serialize(  # type: ignore[override]
        self,
        out: IO[bytes],
        /,
        *,
        stream: Stream | None = None,
        options: SerializerOptions | None = None,
        **unused: Any,
    ) -> None:
        """
        Serialize self.store content to Jelly format.

        Args:
            out (IO[bytes]): output buffered writer
            stream (Stream | None, optional): Jelly stream object. Defaults to None.
            options (SerializerOptions | None, optional): Serializer options
                if defined beforehand, e.g., read from a separate file.
                Defaults to None.
            **unused(Any): unused args for RDFLib serialize

        """
        if options is None:
            options = guess_options(self.store)
        if stream is None:
            stream = guess_stream(options, self.store)
        write = write_delimited if stream.options.params.delimited else write_single
        for stream_frame in stream_frames(stream, self.store):
            write(stream_frame, out)


def grouped_stream_to_frames(
    sink_generator: Generator[Graph] | Generator[Dataset],
    options: SerializerOptions | None = None,
) -> Generator[jelly.RdfStreamFrame]:
    """
    Transform Graphs/Datasets into Jelly frames, one frame per Graph/Dataset.

    Note: options are guessed if not provided.

    Args:
        sink_generator (Generator[Graph] | Generator[Dataset]): Generator of
            Graphs/Dataset to transform.
        options (SerializerOptions | None, optional): stream options to use.
            Options are guessed based on the sink store type. Defaults to None.

    Yields:
        Generator[jelly.RdfStreamFrame]: produced Jelly frames

    """
    stream = None
    for sink in sink_generator:
        if not stream:
            if options is None:
                options = guess_options(sink)
            stream = guess_stream(options, sink)
        yield from stream_frames(stream, sink)


def grouped_stream_to_file(
    stream: Generator[Graph] | Generator[Dataset],
    output_file: IO[bytes],
    **kwargs: Any,
) -> None:
    """
    Write stream of Graphs/Datasets to a binary file.

    Args:
        stream (Generator[Graph] | Generator[Dataset]): Generator of
            Graphs/Dataset to transform.
        output_file (IO[bytes]): output buffered writer.
        **kwargs (Any): options to pass to stream.

    """
    for frame in grouped_stream_to_frames(stream, **kwargs):
        write_delimited(frame, output_file)


def flat_stream_to_frames(
    statements: Generator[Triple | Quad],
    options: SerializerOptions | None = None,
) -> Generator[jelly.RdfStreamFrame]:
    """
    Serialize a stream of raw triples or quads into Jelly frames.

    Args:
        statements (Generator[Triple | Quad]):
          s/p/o triples or s/p/o/g quads to serialize.
        options (SerializerOptions | None, optional):
            if omitted, guessed based on the first tuple.

    Yields:
        Generator[jelly.RdfStreamFrame]: generated frames.

    """
    first = next(statements, None)
    if first is None:
        return

    sink = Dataset() if len(first) == QUAD_ARITY else Graph()
    if options is None:
        options = guess_options(sink)
    stream = guess_stream(options, sink)

    combined: Generator[Triple | Quad] | Graph = (
        item for item in chain([first], statements)
    )

    yield from stream_frames(stream, combined)


def flat_stream_to_file(
    statements: Generator[Triple | Quad],
    output_file: IO[bytes],
    options: SerializerOptions | None = None,
) -> None:
    """
    Write Triple or Quad events to a binary file in Jelly flat format.

    Args:
        statements (Generator[Triple | Quad]): statements to serialize.
        output_file (IO[bytes]): output buffered writer.
        options (SerializerOptions | None, optional): stream options.

    """
    for frame in flat_stream_to_frames(statements, options):
        write_delimited(frame, output_file)



import os
from collections.abc import Generator, Iterator
from itertools import chain
from typing import IO

from google.protobuf.proto import parse, parse_length_prefixed

from pyjelly import jelly
from pyjelly.errors import JellyConformanceError
from pyjelly.parse.decode import ParserOptions, options_from_frame


def delimited_jelly_hint(header: bytes) -> bool:
    """
    Detect whether a Jelly file is delimited from its first 3 bytes.

    Truth table (notation: `0A` = `0x0A`, `NN` = `not 0x0A`, `??` = _don't care_):

    | Byte 1 | Byte 2 | Byte 3 | Result                                   |
    |--------|--------|--------|------------------------------------------|
    | `NN`   |  `??`  |  `??`  | Delimited                                |
    | `0A`   |  `NN`  |  `??`  | Non-delimited                            |
    | `0A`   |  `0A`  |  `NN`  | Delimited (size = 10)                    |
    | `0A`   |  `0A`  |  `0A`  | Non-delimited (stream options size = 10) |

    >>> delimited_jelly_hint(bytes([0x00, 0x00, 0x00]))
    True

    >>> delimited_jelly_hint(bytes([0x00, 0x00, 0x0A]))
    True

    >>> delimited_jelly_hint(bytes([0x00, 0x0A, 0x00]))
    True

    >>> delimited_jelly_hint(bytes([0x00, 0x0A, 0x0A]))
    True

    >>> delimited_jelly_hint(bytes([0x0A, 0x00, 0x00]))
    False

    >>> delimited_jelly_hint(bytes([0x0A, 0x00, 0x0A]))
    False

    >>> delimited_jelly_hint(bytes([0x0A, 0x0A, 0x00]))
    True

    >>> delimited_jelly_hint(bytes([0x0A, 0x0A, 0x0A]))
    False
    """
    magic = 0x0A
    return len(header) == 3 and (  # noqa: PLR2004
        header[0] != magic or (header[1] == magic and header[2] != magic)
    )


def frame_iterator(inp: IO[bytes]) -> Generator[jelly.RdfStreamFrame]:
    while frame := parse_length_prefixed(jelly.RdfStreamFrame, inp):
        yield frame


def get_options_and_frames(
    inp: IO[bytes],
) -> tuple[ParserOptions, Iterator[jelly.RdfStreamFrame]]:
    """
    Return stream options and frames from the buffered binary stream.

    Args:
        inp (IO[bytes]): jelly buffered binary stream

    Raises:
        JellyConformanceError: if no non-empty frames detected in the delimited stream
        JellyConformanceError: if non-delimited,
            error is raised if no rows are detected (empty frame)

    Returns:
        tuple[ParserOptions, Iterator[jelly.RdfStreamFrame]]: ParserOptions holds:
            stream types, lookup presets and other stream options

    """
    is_delimited = delimited_jelly_hint(bytes_read := inp.read(3))
    inp.seek(-len(bytes_read), os.SEEK_CUR)

    if is_delimited:
        first_frame = None
        skipped_frames = []
        frames = frame_iterator(inp)
        for frame in frames:
            if not frame.rows:
                skipped_frames.append(frame)
            else:
                first_frame = frame
                break
        if first_frame is None:
            msg = "No non-empty frames found in the stream"
            raise JellyConformanceError(msg)

        options = options_from_frame(first_frame, delimited=True)
        return options, chain(skipped_frames, (first_frame,), frames)

    frame = parse(jelly.RdfStreamFrame, inp.read())

    if not frame.rows:
        msg = "The stream is corrupted (only contains an empty frame)"
        raise JellyConformanceError(msg)

    options = options_from_frame(frame, delimited=False)
    return options, iter((frame,))



from pyjelly.integrations.generic.generic_sink import GenericStatementSink, Triple

# Create a generic sink object, identifier is optional
generic_sink = GenericStatementSink(identifier="my_graph")

# Load triples from the Jelly file
with open("output.jelly", "rb") as in_file:
    generic_sink.parse(in_file)

# Let's inspect them statement by statement
for statement in generic_sink:
    if isinstance(statement, Triple):
        print(statement)

print("All done.")



from __future__ import annotations

import pytest

from pyjelly import jelly
from pyjelly.errors import JellyAssertionError
from pyjelly.options import StreamParameters
from pyjelly.serialize.encode import TermEncoder
from pyjelly.serialize.flows import (
    DatasetsFrameFlow,
    FlatQuadsFrameFlow,
    FlatTriplesFrameFlow,
    FrameFlow,
    GraphsFrameFlow,
    ManualFrameFlow,
    flow_for_type,
)
from pyjelly.serialize.streams import (
    GraphStream,
    QuadStream,
    SerializerOptions,
    Stream,
    TripleStream,
)


def test_flat_triples_inference_delimited() -> None:
    stream = TripleStream(encoder=TermEncoder(), options=None)
    assert isinstance(stream.flow, FlatTriplesFrameFlow)
    assert stream.flow.logical_type == jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES


@pytest.mark.parametrize("stream_class", [QuadStream, GraphStream])
def test_flat_quads_inference_delimited(
    stream_class: type[GraphStream | QuadStream],
) -> None:
    stream = stream_class(encoder=TermEncoder(), options=None)
    assert isinstance(stream.flow, FlatQuadsFrameFlow)
    assert stream.flow.logical_type == jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS


@pytest.mark.parametrize(
    (
        "logical_type",
        "flow_type",
    ),
    [
        (jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES, FlatTriplesFrameFlow),
        (jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS, FlatQuadsFrameFlow),
        (jelly.LOGICAL_STREAM_TYPE_GRAPHS, GraphsFrameFlow),
        (jelly.LOGICAL_STREAM_TYPE_DATASETS, DatasetsFrameFlow),
        (jelly.LOGICAL_STREAM_TYPE_SUBJECT_GRAPHS, GraphsFrameFlow),
        (jelly.LOGICAL_STREAM_TYPE_NAMED_GRAPHS, DatasetsFrameFlow),
        (jelly.LOGICAL_STREAM_TYPE_TIMESTAMPED_NAMED_GRAPHS, DatasetsFrameFlow),
    ],
)
def test_flow_base_logical_type(
    logical_type: jelly.LogicalStreamType, flow_type: type[FrameFlow]
) -> None:
    assert flow_for_type(logical_type) is flow_type


@pytest.mark.parametrize(
    ("stream_class", "logical_type"),
    [
        (TripleStream, jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES),
        (TripleStream, jelly.LOGICAL_STREAM_TYPE_GRAPHS),
        (QuadStream, jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS),
        (QuadStream, jelly.LOGICAL_STREAM_TYPE_DATASETS),
        (GraphStream, jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS),
        (GraphStream, jelly.LOGICAL_STREAM_TYPE_DATASETS),
    ],
)
def test_flow_logical_type_nondelimited(
    stream_class: type[Stream],
    logical_type: jelly.LogicalStreamType,
) -> None:
    stream = stream_class(
        encoder=TermEncoder(),
        options=SerializerOptions(
            logical_type=logical_type,
            params=StreamParameters(delimited=False),
        ),
    )
    assert isinstance(stream.flow, ManualFrameFlow)
    assert stream.flow.logical_type == logical_type


@pytest.mark.parametrize(
    ("stream_class", "logical_type"),
    [
        (TripleStream, jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS),
        (QuadStream, jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES),
        (GraphStream, jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES),
    ],
)
def test_wrong_inference(
    stream_class: type[Stream], logical_type: jelly.LogicalStreamType
) -> None:
    # A couple obvious examples of incompatible combinations
    with pytest.raises(JellyAssertionError):
        stream_class(
            encoder=TermEncoder(),
            options=SerializerOptions(
                logical_type=logical_type,
                params=StreamParameters(delimited=False),
            ),
        )



# ruff: noqa: ARG002,FBT001,FBT002

from __future__ import annotations

from collections.abc import Generator, Iterator
from itertools import groupby
from typing import Any
from typing_extensions import TypeAlias

from rdflib.graph import Graph
from rdflib.store import Store
from rdflib.term import Identifier, Node, URIRef

Triple: TypeAlias = tuple[Node, Node, Node]


class OrderedMemory(Store):
    context_aware = True
    graph_aware = True

    def __init__(
        self,
        configuration: str | None = None,
        identifier: Identifier | None = None,
    ) -> None:
        super().__init__(configuration=configuration, identifier=identifier)
        self._quads: list[tuple[Triple, Graph | None]] = []
        self._seen_contexts: dict[str, Graph] = {}
        self._namespaces: dict[str, URIRef] = {}
        self._prefixes: dict[URIRef, str] = {}

    def add(
        self, triple: Triple, context: Graph | None = None, quoted: bool = False
    ) -> None:
        """
        Add a triple to existing context (Graph) and adds context to list.

        Args:
            triple (Triple): a tuple of terms
            context (Graph | None, optional): Defaults to None.
            quoted (bool, optional): _description_. Defaults to False.
            TODO: why is quoted arg here?

        """
        self._quads.append((triple, context))
        if context is not None:
            self._seen_contexts[str(context.identifier)] = context

    def triples(
        self, triple_pattern: Any = None, context: Graph | None = None
    ) -> Iterator[tuple[Triple, Iterator[Graph | None]]]:
        for triple, ctx in self._quads:
            if context is None or context == ctx:
                yield triple, iter((ctx,))

    def contexts(self, triple: Triple | None = None) -> Generator[Graph]:
        if triple is None:
            for ctx, statements in groupby(self._quads, key=lambda x: x[1]):
                if ctx is not None:
                    new_store = self.__class__()
                    graph = Graph(store=new_store, identifier=ctx.identifier)
                    for statement, _ in statements:
                        graph.add(statement)
                    yield graph
        else:
            seen_ids = set()
            for t, ctx in self._quads:
                if t == triple and ctx is not None:
                    ctx_id = str(ctx.identifier)
                    if ctx_id not in seen_ids:
                        seen_ids.add(ctx_id)
                        yield ctx

    def add_graph(self, graph: Graph) -> None:
        self._seen_contexts[graph.identifier] = graph

    def remove_graph(self, graph: Graph) -> None:
        self._quads = [(t, ctx) for (t, ctx) in self._quads if ctx != graph]
        self._seen_contexts.pop(graph.identifier, None)

    def __len__(self, context: Graph | None = None) -> int:
        return sum(1 for _, ctx in self._quads if context is None or ctx == context)

    def bind(self, prefix: str, namespace: URIRef, override: bool = True) -> None:
        if override or prefix not in self._namespaces:
            self._namespaces[prefix] = namespace
        if override or namespace not in self._prefixes:
            self._prefixes[namespace] = prefix

    def prefix(self, namespace: URIRef) -> str | None:
        return self._prefixes.get(namespace)

    def namespace(self, prefix: str) -> URIRef | None:
        return self._namespaces.get(prefix)

    def namespaces(self) -> Iterator[tuple[str, URIRef]]:
        return iter(self._namespaces.items())



import pytest
from inline_snapshot import snapshot

from pyjelly.serialize.lookup import Lookup


def current_size(lookup: Lookup) -> int:
    return len(lookup.data)


def test_maintains_size_0() -> None:
    lookup = Lookup(0)
    with pytest.raises(IndexError, match="cannot insert"):
        assert lookup.insert("foo") is None
    assert current_size(lookup) == 0


@pytest.mark.parametrize("max_size", [1, 2, 10, 11, 50, 1190])
def test_maintains_size(max_size: int) -> None:
    lookup = Lookup(max_size)

    # Sequential path
    for i in range(max_size):
        lookup.insert(f"key{i}")
        # It is not important to use exactly as much space as necessary
        # at this point of time. The critical requirement is to never
        # exceed the maximum size
        assert current_size(lookup) <= max_size

    # Evicting path
    for i in range(max_size, max_size * 2):
        lookup.insert(f"key{i}")
        # In the evicting path, it is never possible for the lookup
        # to be at less than its full capacity, because pre-existing entries
        # never expire
        assert current_size(lookup) == max_size


def test_insert_sequential() -> None:
    lookup = Lookup(3)

    assert lookup.insert("key1") == snapshot(1)
    assert lookup.insert("key2") == snapshot(2)
    assert lookup.insert("key3") == snapshot(3)


def test_insert_evicting() -> None:
    lookup = Lookup(3)

    lookup.insert("key1")
    lookup.insert("key2")
    lookup.insert("key3")

    assert lookup.insert("key4") == snapshot(1)
    assert lookup.insert("key5") == snapshot(2)
    assert lookup.insert("key6") == snapshot(3)
    assert lookup.insert("key7") == snapshot(1)
    assert lookup.insert("key8") == snapshot(2)
    assert lookup.insert("key9") == snapshot(3)


def test_insert_asserts_fresh_key() -> None:
    lookup = Lookup(1)
    lookup.insert("foo")
    with pytest.raises(AssertionError, match="key 'foo' already present"):
        lookup.insert("foo")


def test_make_last_to_evict() -> None:
    lookup = Lookup(3)
    a = lookup.insert("a")
    # evict order: a

    b = lookup.insert("b")
    # evict order: a, b

    lookup.make_last_to_evict("a")
    # b, a

    c = lookup.insert("c")
    # b, a, c

    d = lookup.insert("d")  # evicts b
    # a, c, d

    assert d == b

    e = lookup.insert("e")  # evicts a
    # c, d, e

    assert e == a

    f = lookup.insert("f")  # evicts c
    # d, e, f

    assert f == c

    lookup.make_last_to_evict("e")
    # d, f, e

    g = lookup.insert("g")  # evicts d
    # f, e, g

    assert g == d

    h = lookup.insert("h")  # evicts f
    # e, g, h

    assert h == f

    i = lookup.insert("i")  # evicts e
    # g, h, i

    assert i == e


def test_make_last_to_evict_for_existing_key_raises() -> None:
    lookup = Lookup(1)
    with pytest.raises(KeyError, match="key1"):
        lookup.make_last_to_evict("key1")


def test_lookup_repr() -> None:
    lk = Lookup(1)
    lk.insert("a")
    assert str(lk) == f"Lookup(max_size={lk.max_size!r}, data={lk.data!r})"



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: rdf.proto
# Protobuf Python Version: 6.31.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    0,
    '',
    'rdf.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\trdf.proto\x12!eu.ostrzyciel.jelly.core.proto.v1\",\n\x06RdfIri\x12\x11\n\tprefix_id\x18\x01 \x01(\r\x12\x0f\n\x07name_id\x18\x02 \x01(\r\"O\n\nRdfLiteral\x12\x0b\n\x03lex\x18\x01 \x01(\t\x12\x11\n\x07langtag\x18\x02 \x01(\tH\x00\x12\x12\n\x08\x64\x61tatype\x18\x03 \x01(\rH\x00\x42\r\n\x0bliteralKind\"\x11\n\x0fRdfDefaultGraph\"\xbb\x05\n\tRdfTriple\x12:\n\x05s_iri\x18\x01 \x01(\x0b\x32).eu.ostrzyciel.jelly.core.proto.v1.RdfIriH\x00\x12\x11\n\x07s_bnode\x18\x02 \x01(\tH\x00\x12\x42\n\ts_literal\x18\x03 \x01(\x0b\x32-.eu.ostrzyciel.jelly.core.proto.v1.RdfLiteralH\x00\x12\x45\n\rs_triple_term\x18\x04 \x01(\x0b\x32,.eu.ostrzyciel.jelly.core.proto.v1.RdfTripleH\x00\x12:\n\x05p_iri\x18\x05 \x01(\x0b\x32).eu.ostrzyciel.jelly.core.proto.v1.RdfIriH\x01\x12\x11\n\x07p_bnode\x18\x06 \x01(\tH\x01\x12\x42\n\tp_literal\x18\x07 \x01(\x0b\x32-.eu.ostrzyciel.jelly.core.proto.v1.RdfLiteralH\x01\x12\x45\n\rp_triple_term\x18\x08 \x01(\x0b\x32,.eu.ostrzyciel.jelly.core.proto.v1.RdfTripleH\x01\x12:\n\x05o_iri\x18\t \x01(\x0b\x32).eu.ostrzyciel.jelly.core.proto.v1.RdfIriH\x02\x12\x11\n\x07o_bnode\x18\n \x01(\tH\x02\x12\x42\n\to_literal\x18\x0b \x01(\x0b\x32-.eu.ostrzyciel.jelly.core.proto.v1.RdfLiteralH\x02\x12\x45\n\ro_triple_term\x18\x0c \x01(\x0b\x32,.eu.ostrzyciel.jelly.core.proto.v1.RdfTripleH\x02\x42\t\n\x07subjectB\x0b\n\tpredicateB\x08\n\x06object\"\xa4\x07\n\x07RdfQuad\x12:\n\x05s_iri\x18\x01 \x01(\x0b\x32).eu.ostrzyciel.jelly.core.proto.v1.RdfIriH\x00\x12\x11\n\x07s_bnode\x18\x02 \x01(\tH\x00\x12\x42\n\ts_literal\x18\x03 \x01(\x0b\x32-.eu.ostrzyciel.jelly.core.proto.v1.RdfLiteralH\x00\x12\x45\n\rs_triple_term\x18\x04 \x01(\x0b\x32,.eu.ostrzyciel.jelly.core.proto.v1.RdfTripleH\x00\x12:\n\x05p_iri\x18\x05 \x01(\x0b\x32).eu.ostrzyciel.jelly.core.proto.v1.RdfIriH\x01\x12\x11\n\x07p_bnode\x18\x06 \x01(\tH\x01\x12\x42\n\tp_literal\x18\x07 \x01(\x0b\x32-.eu.ostrzyciel.jelly.core.proto.v1.RdfLiteralH\x01\x12\x45\n\rp_triple_term\x18\x08 \x01(\x0b\x32,.eu.ostrzyciel.jelly.core.proto.v1.RdfTripleH\x01\x12:\n\x05o_iri\x18\t \x01(\x0b\x32).eu.ostrzyciel.jelly.core.proto.v1.RdfIriH\x02\x12\x11\n\x07o_bnode\x18\n \x01(\tH\x02\x12\x42\n\to_literal\x18\x0b \x01(\x0b\x32-.eu.ostrzyciel.jelly.core.proto.v1.RdfLiteralH\x02\x12\x45\n\ro_triple_term\x18\x0c \x01(\x0b\x32,.eu.ostrzyciel.jelly.core.proto.v1.RdfTripleH\x02\x12:\n\x05g_iri\x18\r \x01(\x0b\x32).eu.ostrzyciel.jelly.core.proto.v1.RdfIriH\x03\x12\x11\n\x07g_bnode\x18\x0e \x01(\tH\x03\x12M\n\x0fg_default_graph\x18\x0f \x01(\x0b\x32\x32.eu.ostrzyciel.jelly.core.proto.v1.RdfDefaultGraphH\x03\x12\x42\n\tg_literal\x18\x10 \x01(\x0b\x32-.eu.ostrzyciel.jelly.core.proto.v1.RdfLiteralH\x03\x42\t\n\x07subjectB\x0b\n\tpredicateB\x08\n\x06objectB\x07\n\x05graph\"\xfa\x01\n\rRdfGraphStart\x12:\n\x05g_iri\x18\x01 \x01(\x0b\x32).eu.ostrzyciel.jelly.core.proto.v1.RdfIriH\x00\x12\x11\n\x07g_bnode\x18\x02 \x01(\tH\x00\x12M\n\x0fg_default_graph\x18\x03 \x01(\x0b\x32\x32.eu.ostrzyciel.jelly.core.proto.v1.RdfDefaultGraphH\x00\x12\x42\n\tg_literal\x18\x04 \x01(\x0b\x32-.eu.ostrzyciel.jelly.core.proto.v1.RdfLiteralH\x00\x42\x07\n\x05graph\"\r\n\x0bRdfGraphEnd\"a\n\x17RdfNamespaceDeclaration\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x38\n\x05value\x18\x02 \x01(\x0b\x32).eu.ostrzyciel.jelly.core.proto.v1.RdfIri\")\n\x0cRdfNameEntry\x12\n\n\x02id\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\t\"+\n\x0eRdfPrefixEntry\x12\n\n\x02id\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\t\"-\n\x10RdfDatatypeEntry\x12\n\n\x02id\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\t\"\xe1\x02\n\x10RdfStreamOptions\x12\x13\n\x0bstream_name\x18\x01 \x01(\t\x12L\n\rphysical_type\x18\x02 \x01(\x0e\x32\x35.eu.ostrzyciel.jelly.core.proto.v1.PhysicalStreamType\x12\x1e\n\x16generalized_statements\x18\x03 \x01(\x08\x12\x10\n\x08rdf_star\x18\x04 \x01(\x08\x12\x1b\n\x13max_name_table_size\x18\t \x01(\r\x12\x1d\n\x15max_prefix_table_size\x18\n \x01(\r\x12\x1f\n\x17max_datatype_table_size\x18\x0b \x01(\r\x12J\n\x0clogical_type\x18\x0e \x01(\x0e\x32\x34.eu.ostrzyciel.jelly.core.proto.v1.LogicalStreamType\x12\x0f\n\x07version\x18\x0f \x01(\r\"\x87\x05\n\x0cRdfStreamRow\x12\x46\n\x07options\x18\x01 \x01(\x0b\x32\x33.eu.ostrzyciel.jelly.core.proto.v1.RdfStreamOptionsH\x00\x12>\n\x06triple\x18\x02 \x01(\x0b\x32,.eu.ostrzyciel.jelly.core.proto.v1.RdfTripleH\x00\x12:\n\x04quad\x18\x03 \x01(\x0b\x32*.eu.ostrzyciel.jelly.core.proto.v1.RdfQuadH\x00\x12G\n\x0bgraph_start\x18\x04 \x01(\x0b\x32\x30.eu.ostrzyciel.jelly.core.proto.v1.RdfGraphStartH\x00\x12\x43\n\tgraph_end\x18\x05 \x01(\x0b\x32..eu.ostrzyciel.jelly.core.proto.v1.RdfGraphEndH\x00\x12O\n\tnamespace\x18\x06 \x01(\x0b\x32:.eu.ostrzyciel.jelly.core.proto.v1.RdfNamespaceDeclarationH\x00\x12?\n\x04name\x18\t \x01(\x0b\x32/.eu.ostrzyciel.jelly.core.proto.v1.RdfNameEntryH\x00\x12\x43\n\x06prefix\x18\n \x01(\x0b\x32\x31.eu.ostrzyciel.jelly.core.proto.v1.RdfPrefixEntryH\x00\x12G\n\x08\x64\x61tatype\x18\x0b \x01(\x0b\x32\x33.eu.ostrzyciel.jelly.core.proto.v1.RdfDatatypeEntryH\x00\x42\x05\n\x03row\"\xd3\x01\n\x0eRdfStreamFrame\x12=\n\x04rows\x18\x01 \x03(\x0b\x32/.eu.ostrzyciel.jelly.core.proto.v1.RdfStreamRow\x12Q\n\x08metadata\x18\x0f \x03(\x0b\x32?.eu.ostrzyciel.jelly.core.proto.v1.RdfStreamFrame.MetadataEntry\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c:\x02\x38\x01*\x9d\x01\n\x12PhysicalStreamType\x12$\n PHYSICAL_STREAM_TYPE_UNSPECIFIED\x10\x00\x12 \n\x1cPHYSICAL_STREAM_TYPE_TRIPLES\x10\x01\x12\x1e\n\x1aPHYSICAL_STREAM_TYPE_QUADS\x10\x02\x12\x1f\n\x1bPHYSICAL_STREAM_TYPE_GRAPHS\x10\x03*\xc4\x02\n\x11LogicalStreamType\x12#\n\x1fLOGICAL_STREAM_TYPE_UNSPECIFIED\x10\x00\x12$\n LOGICAL_STREAM_TYPE_FLAT_TRIPLES\x10\x01\x12\"\n\x1eLOGICAL_STREAM_TYPE_FLAT_QUADS\x10\x02\x12\x1e\n\x1aLOGICAL_STREAM_TYPE_GRAPHS\x10\x03\x12 \n\x1cLOGICAL_STREAM_TYPE_DATASETS\x10\x04\x12&\n\"LOGICAL_STREAM_TYPE_SUBJECT_GRAPHS\x10\r\x12$\n LOGICAL_STREAM_TYPE_NAMED_GRAPHS\x10\x0e\x12\x30\n,LOGICAL_STREAM_TYPE_TIMESTAMPED_NAMED_GRAPHS\x10rb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'rdf_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_RDFSTREAMFRAME_METADATAENTRY']._loaded_options = None
  _globals['_RDFSTREAMFRAME_METADATAENTRY']._serialized_options = b'8\001'
  _globals['_PHYSICALSTREAMTYPE']._serialized_start=3554
  _globals['_PHYSICALSTREAMTYPE']._serialized_end=3711
  _globals['_LOGICALSTREAMTYPE']._serialized_start=3714
  _globals['_LOGICALSTREAMTYPE']._serialized_end=4038
  _globals['_RDFIRI']._serialized_start=48
  _globals['_RDFIRI']._serialized_end=92
  _globals['_RDFLITERAL']._serialized_start=94
  _globals['_RDFLITERAL']._serialized_end=173
  _globals['_RDFDEFAULTGRAPH']._serialized_start=175
  _globals['_RDFDEFAULTGRAPH']._serialized_end=192
  _globals['_RDFTRIPLE']._serialized_start=195
  _globals['_RDFTRIPLE']._serialized_end=894
  _globals['_RDFQUAD']._serialized_start=897
  _globals['_RDFQUAD']._serialized_end=1829
  _globals['_RDFGRAPHSTART']._serialized_start=1832
  _globals['_RDFGRAPHSTART']._serialized_end=2082
  _globals['_RDFGRAPHEND']._serialized_start=2084
  _globals['_RDFGRAPHEND']._serialized_end=2097
  _globals['_RDFNAMESPACEDECLARATION']._serialized_start=2099
  _globals['_RDFNAMESPACEDECLARATION']._serialized_end=2196
  _globals['_RDFNAMEENTRY']._serialized_start=2198
  _globals['_RDFNAMEENTRY']._serialized_end=2239
  _globals['_RDFPREFIXENTRY']._serialized_start=2241
  _globals['_RDFPREFIXENTRY']._serialized_end=2284
  _globals['_RDFDATATYPEENTRY']._serialized_start=2286
  _globals['_RDFDATATYPEENTRY']._serialized_end=2331
  _globals['_RDFSTREAMOPTIONS']._serialized_start=2334
  _globals['_RDFSTREAMOPTIONS']._serialized_end=2687
  _globals['_RDFSTREAMROW']._serialized_start=2690
  _globals['_RDFSTREAMROW']._serialized_end=3337
  _globals['_RDFSTREAMFRAME']._serialized_start=3340
  _globals['_RDFSTREAMFRAME']._serialized_end=3551
  _globals['_RDFSTREAMFRAME_METADATAENTRY']._serialized_start=3504
  _globals['_RDFSTREAMFRAME_METADATAENTRY']._serialized_end=3551
# @@protoc_insertion_point(module_scope)



import pytest
from rdflib import BNode, Literal, URIRef

from tests.utils.ordered_memory import OrderedMemory, Triple


@pytest.mark.parametrize(
    "triples",
    [
        [
            (
                URIRef("http://foo.com/a"),
                URIRef("http://foo.bar/b"),
                URIRef("http://foo.biz/c"),
            ),
            (
                URIRef("http://bar.spam/x"),
                URIRef("http://foo.eggs/y"),
                URIRef("http://ham.spam/z"),
            ),
        ],
        [
            (
                BNode("a"),
                URIRef("http://example.org/p"),
                Literal("test"),
            ),
            (
                URIRef("http://example.org/s"),
                URIRef("http://example.org/p"),
                Literal(42),
            ),
        ],
        [
            (
                URIRef("http://x"),
                URIRef("http://y"),
                Literal("hello", lang="en"),
            ),
            (
                URIRef("http://x"),
                URIRef("http://y"),
                Literal(
                    "3.14", datatype=URIRef("http://www.w3.org/2001/XMLSchema#decimal")
                ),
            ),
        ],
    ],
)
def test_ordered_memory(triples: list[Triple]) -> None:
    store = OrderedMemory()
    for triple in triples:
        store.add(triple)
    output_triples = [t for t, _ in store.triples()]
    assert output_triples == triples



from typing import IO

from google.protobuf.proto import serialize_length_prefixed

from pyjelly import jelly


def write_delimited(frame: jelly.RdfStreamFrame, output_stream: IO[bytes]) -> None:
    serialize_length_prefixed(frame, output_stream)


def write_single(frame: jelly.RdfStreamFrame, output_stream: IO[bytes]) -> None:
    output_stream.write(frame.SerializeToString(deterministic=True))



from __future__ import annotations

from collections import UserList
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Any, ClassVar
from typing_extensions import override

from pyjelly import jelly

DEFAULT_FRAME_SIZE = 250


class FrameFlow(UserList[jelly.RdfStreamRow]):
    """
    Abstract base class for producing Jelly frames from RDF stream rows.

    Collects stream rows and assembles them into RdfStreamFrame objects when ready.

    Allows for passing LogicalStreamType, required for
        logical subtypes and non-delimited streams.
    """

    logical_type: jelly.LogicalStreamType
    registry: ClassVar[dict[jelly.LogicalStreamType, type[FrameFlow]]] = {}

    def __init__(
        self,
        initlist: Iterable[jelly.RdfStreamRow] | None = None,
        *,
        logical_type: jelly.LogicalStreamType | None = None,
        **__kwargs: Any,
    ) -> None:
        super().__init__(initlist)
        self.logical_type = logical_type or self.__class__.logical_type

    def frame_from_graph(self) -> jelly.RdfStreamFrame | None:
        """
        Treat the current rows as a graph and produce a frame.

        Default implementation returns None.
        """
        return None

    def frame_from_dataset(self) -> jelly.RdfStreamFrame | None:
        """
        Treat the current rows as a dataset and produce a frame.

        Default implementation returns None.
        """
        return None

    def frame_from_bounds(self) -> jelly.RdfStreamFrame | None:
        return None

    def to_stream_frame(self) -> jelly.RdfStreamFrame | None:
        """
        Create stream frame from flow content.

        Notes:
            Clears flow content after creating the frame.

        Returns:
            jelly.RdfStreamFrame | None: stream frame

        """
        if not self:
            return None
        frame = jelly.RdfStreamFrame(rows=self)
        self.clear()
        return frame


class ManualFrameFlow(FrameFlow):
    """
    Produces frames only when manually requested (never automatically).

    !!! warning
        All stream rows are kept in memory until `to_stream_frame()` is called.
        This may lead to high memory usage for large streams.

    Used for non-delimited serialization.
    """

    logical_type = jelly.LOGICAL_STREAM_TYPE_UNSPECIFIED


@dataclass
class BoundedFrameFlow(FrameFlow):
    """
    Produce frames automatically when a fixed number of rows is reached.

    Used for delimited encoding (default mode).
    """

    logical_type = jelly.LOGICAL_STREAM_TYPE_UNSPECIFIED
    frame_size: int

    @override
    def __init__(
        self,
        initlist: Iterable[jelly.RdfStreamRow] | None = None,
        logical_type: jelly.LogicalStreamType | None = None,
        *,
        frame_size: int | None = None,
    ) -> None:
        super().__init__(initlist, logical_type=logical_type)
        self.frame_size = frame_size or DEFAULT_FRAME_SIZE

    @override
    def frame_from_bounds(self) -> jelly.RdfStreamFrame | None:
        """
        Emit frame from flow if full.

        Returns:
            jelly.RdfStreamFrame | None: stream frame

        """
        if len(self) >= self.frame_size:
            return self.to_stream_frame()
        return None


class FlatTriplesFrameFlow(BoundedFrameFlow):
    logical_type = jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES


class FlatQuadsFrameFlow(BoundedFrameFlow):
    logical_type = jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS


class GraphsFrameFlow(FrameFlow):
    logical_type = jelly.LOGICAL_STREAM_TYPE_GRAPHS

    def frame_from_graph(self) -> jelly.RdfStreamFrame | None:
        """
        Emit current flow content (one graph) as jelly frame.

        Returns:
            jelly.RdfStreamFrame | None: jelly frame or none if
                flow is empty.

        """
        return self.to_stream_frame()


class DatasetsFrameFlow(FrameFlow):
    logical_type = jelly.LOGICAL_STREAM_TYPE_DATASETS

    def frame_from_dataset(self) -> jelly.RdfStreamFrame | None:
        """
        Emit current flow content (dataset) as jelly frame.

        Returns:
            jelly.RdfStreamFrame | None: jelly frame or none if
                flow is empty.

        """
        return self.to_stream_frame()


FLOW_DISPATCH: dict[jelly.LogicalStreamType, type[FrameFlow]] = {
    jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES: FlatTriplesFrameFlow,
    jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS: FlatQuadsFrameFlow,
    jelly.LOGICAL_STREAM_TYPE_GRAPHS: GraphsFrameFlow,
    jelly.LOGICAL_STREAM_TYPE_DATASETS: DatasetsFrameFlow,
}


def flow_for_type(logical_type: jelly.LogicalStreamType) -> type[FrameFlow]:
    """
    Return flow based on logical type requested.

    Note: uses base logical type for subtypes (i.e., SUBJECT_GRAPHS uses
        the same flow as its base type GRAPHS).

    Args:
        logical_type (jelly.LogicalStreamType): logical type requested.

    Raises:
        NotImplementedError: if (base) logical stream type is not supported.

    Returns:
        type[FrameFlow]: FrameFlow for respective logical type.

    """
    try:
        base_logical_type_value = logical_type % 10
        base_name = jelly.LogicalStreamType.Name(base_logical_type_value)
        return FLOW_DISPATCH[getattr(jelly.LogicalStreamType, base_name)]
    except KeyError:
        msg = (
            "unsupported logical stream type: "
            f"{jelly.LogicalStreamType.Name(logical_type)}"
        )
        raise NotImplementedError(msg) from None



from dataclasses import replace
from itertools import chain, product
from pathlib import Path

import pytest

from pyjelly.options import LookupPreset
from tests.e2e_tests.ser_des.base_ser_des import BaseSerDes
from tests.e2e_tests.ser_des.rdflib_ser_des import RdflibSerDes


class End2EndOptionSetup:
    """Set up stream options, file size and file name for E2E tests."""

    test_root: Path = Path("tests/e2e_test_cases/")

    def setup_ser_des(self) -> list[tuple[BaseSerDes, BaseSerDes, LookupPreset, int]]:
        """Set up test serializer, deserializer, options and frame_size."""
        ser = [RdflibSerDes()]
        des = [RdflibSerDes()]
        # We want to have a variety of options to test
        # Particularly examples of small lookup sizes
        # and a lack of prefix
        small = LookupPreset.small()
        no_prefixes = replace(LookupPreset.small(), max_prefixes=0)
        tiny_lookups = replace(LookupPreset.small(), max_names=16, max_prefixes=8)
        big = LookupPreset()
        presets = [small, no_prefixes, tiny_lookups, big]
        frame_sizes = [1, 4, 200, 10_000]
        return list(product(ser, des, presets, frame_sizes))

    def setup_triple_files(
        self,
    ) -> list[tuple[BaseSerDes, BaseSerDes, LookupPreset, int, Path]]:
        """Set up options for each of the test triple files."""
        test_dir: Path = self.test_root / "triples_rdf_1_1"
        files = test_dir.glob("*.nt")
        options = self.setup_ser_des()
        return list(chain(*[[(*o, f) for o in options] for f in files]))

    def setup_quad_files(
        self,
    ) -> list[tuple[BaseSerDes, BaseSerDes, LookupPreset, int, Path]]:
        """Set up options for each of the test quad files."""
        test_dir: Path = self.test_root / "quads_rdf_1_1"
        files = test_dir.glob("*.nq")
        options = self.setup_ser_des()
        return list(chain(*[[(*o, f) for o in options] for f in files]))


class TestEnd2End:
    setup = End2EndOptionSetup()

    @pytest.mark.parametrize(
        ("ser", "des", "preset", "frame_size", "file"), setup.setup_triple_files()
    )
    def test_triple_files(
        self,
        ser: BaseSerDes,
        des: BaseSerDes,
        preset: LookupPreset,
        frame_size: int,
        file: Path,
    ) -> None:
        nt_reader = RdflibSerDes()
        with file.open("rb") as f:
            triples = nt_reader.read_triples(f.read())
            jelly_io = ser.write_triples_jelly(triples, preset, frame_size)
            new_g = des.read_triples_jelly(jelly_io)
            assert set(triples) == set(new_g)

    @pytest.mark.parametrize(
        ("ser", "des", "preset", "frame_size", "file"), setup.setup_quad_files()
    )
    def test_quad_files(
        self,
        ser: BaseSerDes,
        des: BaseSerDes,
        preset: LookupPreset,
        frame_size: int,
        file: Path,
    ) -> None:
        nq_reader = RdflibSerDes()
        with file.open("rb") as f:
            quads = nq_reader.read_quads(f.read())
            jelly_io = ser.write_quads_jelly(quads, preset, frame_size)
            new_g = des.read_quads_jelly(jelly_io)
            assert set(quads) == set(new_g)



class JellyConformanceError(Exception):
    """Raised when Jelly conformance is violated."""


class JellyAssertionError(AssertionError):
    """Raised when a recommended assertion from the specification fails."""


class JellyNotImplementedError(NotImplementedError):
    """Raised when a future feature is not yet implemented."""



import io
from typing import Any
from unittest.mock import MagicMock, patch

import pytest
from rdflib import Graph
from rdflib import plugin as rdflib_plugin
from rdflib.parser import Parser as RDFLibParser
from rdflib.serializer import Serializer as RDFLibSerializer

from pyjelly import options

rdflib_entrypoint_names = ("jelly", *options.MIMETYPES)
all_entrypoints = pytest.mark.parametrize("file_format", rdflib_entrypoint_names)


def clean_plugin_cache(plugin_name: str, extension_class: type[Any]) -> None:
    # RDFLib caches plugin classes, so a mock is unable to work
    # if the plugin class was already used by some other test.
    plugin = rdflib_plugin._plugins[(plugin_name, extension_class)]
    plugin._class = None


@all_entrypoints
@patch("pyjelly.integrations.rdflib.serialize.RDFLibJellySerializer")
def test_jelly_serializer_discovered(mock: MagicMock, file_format: str) -> None:
    clean_plugin_cache(file_format, RDFLibSerializer)
    try:
        graph = Graph()
        graph.serialize(format=file_format)
        mock.assert_called_once_with(graph)
        mock.return_value.serialize.assert_called_once()
    finally:
        # Don't leak through cache
        clean_plugin_cache(file_format, RDFLibSerializer)


@all_entrypoints
@patch("pyjelly.integrations.rdflib.parse.RDFLibJellyParser")
def test_jelly_parser_discovered(mock: MagicMock, file_format: str) -> None:
    clean_plugin_cache(file_format, RDFLibParser)
    try:
        graph = Graph()
        graph.parse(io.StringIO(), format=file_format)
        mock.assert_called_once_with()
        mock.return_value.parse.assert_called_once()
    finally:
        # Don't leak through cache
        clean_plugin_cache(file_format, RDFLibParser)
